/**
 * Program to test and benchmark clSemiLinkedList.
 */
#include <stdio.h>
#include "../test/cl.h"
#include "../kma.h"
#include "clSemiLinkedList.h"
#include "LinkedList.h"
#include "../simple-opencl/simpleCL.h"
#include<sys/time.h>
#include <math.h>

void readBackList(cl_device_id, cl_context, cl_command_queue,
		cl_program , cl_mem*, cl_mem*);

/* Subtract the struct timeval values X and Y,
 storing the result in RESULT.
 Return 1 if the difference is negative, otherwise 0. */
int timeval_subtract(result, x, y)
	struct timeval *result, *x, *y; {
	/* Perform the carry for the later subtraction by updating y. */
	if (x->tv_usec < y->tv_usec) {
		int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		y->tv_usec -= 1000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000) {
		int nsec = (x->tv_usec - y->tv_usec) / 1000000;
		y->tv_usec += 1000000 * nsec;
		y->tv_sec -= nsec;
	}

	/* Compute the time remaining to wait.
	 tv_usec is certainly positive. */
	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_usec = x->tv_usec - y->tv_usec;

	/* Return 1 if result is negative. */
	return x->tv_sec < y->tv_sec;
}

/**
 * Return the time, in usec, the given event took to finish in given command queue.
 */
cl_ulong getEventTime(cl_command_queue cq, cl_event event) {
	cl_ulong elapsedTime, startTime, endTime;
	cl_int error;
	error = clFinish(cq);
	if (error != CL_SUCCESS) {
		printf("getEventTime: Failed to finish event: %i\n", error);
		sclPrintErrorFlags(error);
		return 0;
	}
	clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START, sizeof(cl_ulong),
			&startTime, NULL);
	clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END, sizeof(cl_ulong),
			&endTime, NULL);
	elapsedTime = endTime - startTime;
//	printf("start: %lu, end: %lu: elapsed %lu \n", startTime, endTime, elapsedTime);
	return elapsedTime;
}

/**
 * Print info for OpenCL device.
 */
void 	getDeviceInfo(cl_device_id dev){
	size_t ret, workGroupSize;
char name[100];
cl_ulong memSize, maxAllocSize;
cl_uint clock, cores;
clGetDeviceInfo(dev, CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(cl_ulong), &memSize, &ret);
clGetDeviceInfo(dev, CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(cl_uint), &clock, &ret);
clGetDeviceInfo(dev, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(cl_uint), &cores, &ret);
clGetDeviceInfo(dev, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(size_t), &workGroupSize, &ret);
clGetDeviceInfo(dev, CL_DEVICE_NAME, 100*sizeof(char), &name, &ret);
clGetDeviceInfo(dev, CL_DEVICE_MAX_MEM_ALLOC_SIZE, sizeof(cl_ulong), &maxAllocSize, &ret);

printf("%s\n", name);
printf("global mem size: %lu\n", memSize);
printf("maximum mem_buffer allocation size: %lu\n", maxAllocSize);
printf("Cores available: %i\n", cores);
printf("Maximum workgroup size: %i\n", workGroupSize);
printf("Max clock frequency: %i\n", clock);
}

/**
 * Concurrently add keys to the list.
 */
int addKeys_GPU(cl_device_id dev, cl_context ctx, cl_command_queue cq,
		cl_program prg, int *keys, int len, cl_mem *heap, cl_mem *list, cl_event *event) {
	cl_int error;
	cl_kernel kernel;
	size_t threads = len;
	int result;
	cl_mem input = clCreateBuffer(ctx,
		CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, threads * sizeof(int), keys,
				&error);
		if(error == CL_SUCCESS){
		kernel = clCreateKernel(prg, "clSemiLinkedList_addKey", &error);
		if(error == CL_SUCCESS){
		clSetKernelArg(kernel, 0, sizeof(cl_mem), heap);
		clSetKernelArg(kernel, 1, sizeof(cl_mem), list);
		clSetKernelArg(kernel, 2, sizeof(cl_mem), &input);
		error = clEnqueueNDRangeKernel(cq, kernel, 1, NULL, &threads, NULL, 0, NULL,
				event);
		if(error != CL_SUCCESS){
			printf("Error enqueueing kernel: %i\n", error);
			sclPrintErrorFlags(error);
		}
		result = 1;
	//	readBackList(dev, ctx, cq, prg, heap, list);
		clReleaseKernel(kernel);
		}else{
			printf("time_GPU: Error creating kernel: %i\n", error);
		}
		clReleaseMemObject(input);
		}else{
			printf("Time_GPU: Error creating input buffer: %i\n", error);
		}
		return result;
}

/**
 * Count the number of items in the list.
 */
int countList(cl_device_id dev, cl_context ctx, cl_command_queue cq,
		cl_program prg, cl_mem *heap, cl_mem *list) {
	cl_int error;
	size_t threads = 1;
	int result = -1;
	cl_kernel kernel = clCreateKernel(prg, "clSemiLinkedList_count", &error);
	if(error == CL_SUCCESS){
cl_mem output = clCreateBuffer(ctx, CL_MEM_WRITE_ONLY, sizeof(int), NULL, &error);
if(error == CL_SUCCESS){
	clSetKernelArg(kernel, 0, sizeof(cl_mem), heap);
			clSetKernelArg(kernel, 1, sizeof(cl_mem), list);
			clSetKernelArg(kernel, 2, sizeof(cl_mem), &output);
			error = clEnqueueNDRangeKernel(cq, kernel, 1, NULL, &threads, NULL, 0, NULL,
							NULL);
			clFinish(cq);
	if(error == CL_SUCCESS){
		error = clEnqueueReadBuffer(cq, output, CL_TRUE, 0, sizeof(int), &result, 0, 0, 0);
		clFinish(cq);
	}
	clReleaseMemObject(output);
}
clReleaseKernel(kernel);
	}
	return result;
}

/**
 * Delete items with given keys from the list. Keys must be distinct. Only the first occurance of a key will be deleted.
 */
int deleteKeys_GPU(cl_device_id dev, cl_context ctx, cl_command_queue cq,
		cl_program prg, int *keys, int len, cl_mem *heap, cl_mem *list, cl_event *event) {
	cl_int error;
	cl_kernel kernel;
	size_t threads = len;
	int result;
	cl_mem input = clCreateBuffer(ctx,
		CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, threads * sizeof(int), keys,
				&error);
		if(error == CL_SUCCESS){
		kernel = clCreateKernel(prg, "clSemiLinkedList_deleteKey", &error);
		if(error == CL_SUCCESS){
		clSetKernelArg(kernel, 0, sizeof(cl_mem), heap);
		clSetKernelArg(kernel, 1, sizeof(cl_mem), list);
		clSetKernelArg(kernel, 2, sizeof(cl_mem), &input);
		error = clEnqueueNDRangeKernel(cq, kernel, 1, NULL, &threads, NULL, 0, NULL,
				event);
		if(error != CL_SUCCESS){
			printf("Error enqueueing kernel: %i\n", error);
			sclPrintErrorFlags(error);
		}
		result = 1;
		clReleaseKernel(kernel);
		}
		clReleaseMemObject(input);
		}
		return result;
}

/**
 * Measure the time it takes to build a list with the given keys.
 */
long time_buildList_GPU(cl_device_id dev, cl_context ctx, cl_command_queue cq,
		cl_program prg, int *keys, int len) {
	long result =0;
	cl_event event;
	cl_int error;
	cl_kernel kernel;
	size_t threads = len;
	cl_mem heap, list;
	/* Create heap with default (sufficient) size */
	heap = kma_create(dev, ctx, cq, prg, 64);
	if(heap == (cl_mem)0){
		printf("time_GPU: failed to initialize heap: NULL\n");
		return 0;
	}
	list = clSemiLinkedList_create(dev, ctx, cq, prg);
	if(list != (cl_mem)0){
	addKeys_GPU(dev, ctx, cq, prg, keys, threads, &heap, &list, &event);
		result = getEventTime(cq, event);
	clReleaseMemObject(list);
	}else{
		printf("time_GPU: could not initialize list: NULL\n");
	}
	clReleaseMemObject(heap);
	//* Convert from nano sec to usec. */
	return result/1000;
}

/**
 * Measure the time it takes to build a list with the given keys.
 */
long time_buildList_CPU(int* keys, int len){
	int num=4096;
	struct timeval tStart, tEnd, tDuration;
LinkedList **list;
long time = 0;
int i = 0;
num = (num/len);
if(num==0){
	num = 1;
}
list = malloc(num*sizeof(LinkedList *));
for(i=0; i<num; i++){
	list[i] = linkedList_create();
}
gettimeofday(&tStart, NULL);
for(i=0; i<num; i++){
linkedList_addKeys(list[i], keys, len);
//linkedList_printList(list);
}
gettimeofday(&tEnd, NULL);
timeval_subtract(&tDuration, &tEnd, &tStart);
time  = tDuration.tv_usec;
time += (tDuration.tv_sec*1000000);
for(i=0; i<num; i++){
	linkedList_destroyList(list[i]);
}
free(list);
return time/num;
}

/**
 * Measure the average time it takes to build a list on GPU.
 */
void time_av_addKey_GPU(cl_device_id dev, cl_context ctx, cl_command_queue cq,
		cl_program prg, int iters, int max, int stepSize) {
int i, j, nodes = stepSize;
int iMax = max/stepSize+1;
long results[iMax][iters];
int keys[max];
printf("nodes, time GPU (usec)\n");
for(i=1; i<iMax; i++){
	long sum = 0;
for(j=0; j<iters; j++){
	results[i][j] = time_buildList_GPU(dev, ctx, cq, prg, &keys, nodes);
//	results[i][j] = time_addKeys_CPU(keys, nodes);
	sum += results[i][j];
	if(results[i][j] == 0){
		/* failed to build list; abort */
		return;
	}
}
printf("%i, %li\n", nodes, sum/iters);
nodes += stepSize;
}
}

/**
 * Measure the average time it takes to build a list on CPU.
 */
void time_av_addKey_CPU(int iters, int max, int stepSize) {
int i, j, nodes = stepSize;
int iMax = max/stepSize+1;
long results[iMax][iters];
int keys[max];
printf("nodes, time CPU (usec)\n");
for(i=1; i<iMax; i++){
	long sum = 0;
for(j=0; j<iters; j++){
	results[i][j] = time_buildList_CPU(keys, nodes);
	sum += results[i][j];
}
printf("%i, %li\n", nodes, sum/iters);
nodes += stepSize;
}
}

/**
 * Test the workings of the semi-linked list.
 * First concurrently add distinct items and count the length of the resulting list. Then concurrently remove all but the first and the last item, and check if the length of the list is 2 (and so the last item did not drop out of the list).
 */
void testList(cl_device_id dev, cl_context ctx, cl_command_queue cq,
		cl_program prg) {
	cl_event event;
	cl_int error;
	int *keys;
	int i;
	size_t threads = 1200;
	int count = 0;
	cl_mem heap, list;
	keys = (int *)malloc(threads*sizeof(int));
	for(i=0; i<threads; i++){
		keys[i] = i;
	}
	heap = kma_create(dev, ctx, cq, prg, 64);
	if(heap != (cl_mem)0){
list = clSemiLinkedList_create(dev, ctx, cq, prg);
if(list!=NULL){
if(addKeys_GPU(dev, ctx, cq, prg, keys, threads, &heap, &list, &event)){
	clFinish(cq);
count = countList(dev, ctx, cq, prg, &heap, &list);
if(count == threads){
	threads-=2;
deleteKeys_GPU(dev, ctx, cq, prg, &keys[1], threads, &heap, &list, &event);
clFinish(cq);
count = countList(dev, ctx, cq, prg, &heap, &list);
clFinish(cq);
if(count==2){
	printf("test passed\n");
	readBackList(dev, ctx, cq, prg, &heap, &list);
}else{
	printf("clSemiLinkedList_test: deleting keys: expected count = %i, but was %i\n", 2, count);
}
}else{
	printf("clSemiLinkedList_test: adding keys: expected %i, but was %i\n", threads, count);
}
}else{
	printf("clSemiLinkedList_test: failed to add keys.\n");
}
	clReleaseMemObject(list);
}
clReleaseMemObject(heap);
}
	free(keys);
}

/**
 * Read the content of a list back in an array.
 */
void readBackList(cl_device_id dev, cl_context ctx, cl_command_queue cq,
		cl_program prg, cl_mem *heap, cl_mem *list) {
	int arraySize = 100;
	int i;
	cl_mem outputbuffer, arraySizeMem;
	cl_int error;
	cl_kernel kernel2;
	size_t threads = 1;
	int result[arraySize];
	for (i = 0; i < arraySize; i++)
		result[i] = 0;
	kernel2 = clCreateKernel(prg, "clSemiLinkedList_toarray", &error);
	if (error != CL_SUCCESS) {
		printf("clSemiLinkedList, could not build 2array kernel: %i\n", error);
		return;
	}
	arraySizeMem = clCreateBuffer(ctx,
	CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, sizeof(int), &arraySize, &error);
	if (error != CL_SUCCESS) {
		printf("clSemiLinkedList: Could not create buffer for array size: %i\n",
				error);
		return;
	}
	outputbuffer = clCreateBuffer(ctx, CL_MEM_WRITE_ONLY,
			arraySize * sizeof(int),
			NULL, &error);
	if (error != CL_SUCCESS) {
		printf("clSemiLinkedList: Could not create array buffer: %i\n", error);
		return;
	}
	clSetKernelArg(kernel2, 0, sizeof(cl_mem), list);
	clSetKernelArg(kernel2, 1, sizeof(cl_mem), heap);
	clSetKernelArg(kernel2, 2, sizeof(cl_mem), &outputbuffer);
	clSetKernelArg(kernel2, 3, sizeof(cl_mem), &arraySizeMem);
	error = clEnqueueNDRangeKernel(cq, kernel2, 1, NULL, &threads, NULL, 0,
	NULL, NULL);
	if (error != CL_SUCCESS) {
		printf("clSemiLinkedList: Could not execute 2array kernel: %i\n",
				error);
		return;
	}
	error = clFinish(cq);
	if (error != CL_SUCCESS) {
		printf("clSemiLinkedList: failed to finish test: %i\n", error);
		return;
	}
	error = clEnqueueReadBuffer(cq, outputbuffer, CL_TRUE, 0,
			arraySize * sizeof(int), result, 0, 0, 0);
	if (error != CL_SUCCESS) {
		printf(
				"clSemiLinkedList: could not read array results back to host: %i\n",
				error);
		return;
	}
	error = clFinish(cq);
	if (error != CL_SUCCESS) {
		printf("clSemiLinkedList: failed to finish readback result: %i\n",
				error);
		return;
	}
	for (i = 0; i < arraySize; i++) {
		printf("%i, ", result[i]);
	}
	printf("\n");
	clReleaseMemObject(outputbuffer);
	clReleaseMemObject(arraySizeMem);
	clReleaseKernel(kernel2);
}

int main(int argc, char** argv) {
	int hw = 0;  // Choose first GPU device.
	sclHard* hardware;
	cl_command_queue cq;
	cl_int error;
	   int found;
char *src[3];
cl_program prg;
	   /* get the hardware */
	   hardware = sclGetAllHardware(&found);
if(found<=0){
	printf("Error: could not get hardware");
	return 1;
}
/* create command queue for event profiling */
cq = clCreateCommandQueue(hardware[hw].context, hardware[hw].device, CL_QUEUE_PROFILING_ENABLE, &error);
/* Compile the program! */
src[0] = kernel_read("clIndexedQueue.cl");
src[1] = kernel_read("kma.cl");
src[2] = kernel_read("linkedlist/clSemiLinkedList.cl");

prg = program_compile(hardware[hw].platform, hardware[hw].context, &(hardware[hw].device), 3, src);

getDeviceInfo(hardware[hw].device);
testList(hardware[hw].device, hardware[hw].context, cq, prg);
time_av_addKey_GPU(hardware[hw].device, hardware[hw].context, cq, prg, 5, 4096, 128);
time_av_addKey_CPU(5, 4096, 128);

free(src[0]);
free(src[1]);
free(src[2]);
}
